/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
 *  SPDX-License-Identifier: GPL-2.0-or-later
 *  For more information, see <https://www.knot-dns.cz/>
 */

#pragma once

#include "knot/include/module.h"

// Forward declarations of static modules (generated by configure).
extern const knotd_mod_api_t knotd_mod_api_alias;
extern const knotd_mod_api_t knotd_mod_api_authsignal;
extern const knotd_mod_api_t knotd_mod_api_cookies;
extern const knotd_mod_api_t knotd_mod_api_dnserr;
extern const knotd_mod_api_t knotd_mod_api_dnsproxy;
extern const knotd_mod_api_t knotd_mod_api_geoip;
extern const knotd_mod_api_t knotd_mod_api_noudp;
extern const knotd_mod_api_t knotd_mod_api_onlinesign;
extern const knotd_mod_api_t knotd_mod_api_probe;
extern const knotd_mod_api_t knotd_mod_api_queryacl;
extern const knotd_mod_api_t knotd_mod_api_rrl;
extern const knotd_mod_api_t knotd_mod_api_stats;
extern const knotd_mod_api_t knotd_mod_api_synthrecord;
extern const knotd_mod_api_t knotd_mod_api_whoami;

// STATIC_MODULES initializer (generated by configure).
#define STATIC_MODULES_INIT \
	{ &knotd_mod_api_alias },\
	{ &knotd_mod_api_authsignal },\
	{ &knotd_mod_api_cookies },\
	{ &knotd_mod_api_dnserr },\
	{ &knotd_mod_api_dnsproxy },\
	{ &knotd_mod_api_geoip },\
	{ &knotd_mod_api_noudp },\
	{ &knotd_mod_api_onlinesign },\
	{ &knotd_mod_api_probe },\
	{ &knotd_mod_api_queryacl },\
	{ &knotd_mod_api_rrl },\
	{ &knotd_mod_api_stats },\
	{ &knotd_mod_api_synthrecord },\
	{ &knotd_mod_api_whoami },
