/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
 *  SPDX-License-Identifier: GPL-2.0-or-later
 *  For more information, see <https://www.knot-dns.cz/>
 */

extern unsigned char bpf_kernel_o[];
extern unsigned int bpf_kernel_o_len;
